% NOTE: You need Matlab R2016b version to run the OpenLab simulator

clear variables;
clc;
close all;

% Add to the matlab path the folder where this file is located and all the
% subfolders
addpath(genpath(pwd));

IdentityServerURL = 'https://live.openlab.app/';

[username, api_key, license_guid] = GetLoginData();

% Select the configuration you want to run simulations on
ConfigurationName = 'Your configuration name';

% Choose a name on your simulation and set some initial parameters
SimulationName = 'Drilling sample (Matlab)';
InitialBitDepth = 2500; % [m]
InitialTopOfStringPosition = 20;

UseTransientMechanicalModel = false;
StepDuration = 1.0;

% Create simulation object
Sim = OpenLabClient(IdentityServerURL, username,api_key,license_guid,ConfigurationName,SimulationName,InitialBitDepth, InitialTopOfStringPosition);

% Time steps to simulate
MaxTimeSteps = 3000;

plotFigures = false;

if plotFigures
    % Create animations
    FlowRateFigure = figure('name', 'Flow rates');
    xlabel('Time [sec]');
    ylabel('Flow rate [L/min]');
    grid
    FlowRateInLine = animatedline(0,0,'Color','b','linewidth',1.5,'Marker','none');
    ChokePumpFlowRateInLine = animatedline(0,0,'Color','g','linewidth',1.5,'Marker','none');
    FlowRateOutLine = animatedline(0,0,'Color','r','linewidth',1.5,'Marker','none');
    legend({'Main pump flow rate in' 'Choke pump flow rate in' 'Flow rate out'})
    
    BHPFigure = figure('name', 'Bottom hole pressure');
    xlabel('Time [sec]');
    ylabel('Pressure [bar]');
    grid
    BHPLine = animatedline(0,0,'Color','b','linewidth',1.5,'Marker','none');
    ControllerReferencePressureLine = animatedline(0,0,'Color','g','linewidth',1.5,'Marker','none');
    legend({'Bottom hole pressure' 'Controller setpoint'})
    
    SPPFigure = figure('name', 'Stand pipe pressure');
    xlabel('Time [sec]');
    ylabel('Pressure [bar]');
    grid
    SPPLine = animatedline(0,0,'Color','b','linewidth',1.5,'Marker','none');
    legend({'Stand pipe pressure'});
    
    ChokePressureFigure = figure('name', 'Choke pressure');
    xlabel('Time [sec]');
    ylabel('Pressure [bar]');
    grid
    ChokePLine = animatedline(0,0,'Color','b','linewidth',1.5,'Marker','none');
    legend({'Choke pressure'});
    
    ChokeOpeningFigure = figure('name', 'Choke opening');
    xlabel('Time [sec]');
    ylabel('Choke opening [%]');
    grid
    ChokeOpeningActLine = animatedline(0,0,'Color','b','linewidth',1.5,'Marker','none');
    ChokeOpeningCmdLine = animatedline(0,0,'Color','g','linewidth',1.5,'Marker','none');
    legend({'Actual opening' 'Setpoint'});
    
    AnnulusPressureFigure = figure('name', 'Annulus pressure profile');
    AnnulusPressurePlot = plot(0,0,'b', 'linewidth',1.5);
    ylabel('MD [m]');
    xlabel('Pressure [bar]');
    legend({'Annulus pressure profile'});
    grid

    AnnulusTemperatureFigure = figure('name', 'Annulus temperature profile');
    AnnulusTemperaturePlot = plot(0,0,'b', 'linewidth',1.5);
    ylabel('MD [m]');
    xlabel('Temperature [K]');
    legend({'Annulus temperature profile'});
    grid
    arrangefigures(1,3/4,true);
end

% Initial setpoints
FlowRateIn = 0;   %  [m3/sec]
TopOfStringVelocity = 0;  %  [m/sec]
ROP = 0;  %  [m/sec]
SurfaceRPM = 0;  %  [revolutions per sec.]
ChokeOpening = 1;  %   [0-1]
ChokePumpFlowRateIn = 0;  %  [m3/sec]
DensityIn = nan;  %  [kg/m3]
BopChokeOpening = 1;   % [0,1]
WOBAutoDriller = false;
DesiredWOB = 0;  %  [kg]
WOBProportionalGain = 0.000001;
WOBIntegralGain = 1e-8;

if Sim.IsOK
    try
        for timeStep = 1 : MaxTimeSteps
            tStartStep = tic;
                       
            % Set setpoints to the simulator            
            if timeStep < 150
                ROP = 0/3600;  %  [m/sec]
                TopOfStringVelocity = 0.1;  %  [m/sec]
                FlowRateIn = 1000/60000;
                SurfaceRPM = 120/60;  %  [revolutions per sec.]               
            elseif timeStep < 600
                FlowRateIn = 2200/60000;  %  [m3/sec]
                TopOfStringVelocity = 0.0;  %  [m/sec]
                WOBAutoDriller = true;
                DesiredWOB = 5000;  %  [kg]
                WOBProportionalGain = 1e-5;
                WOBIntegralGain = 1e-6;
            elseif timeStep < 1000
                FlowRateIn = 2200/60000;  %  [m3/sec]
                TopOfStringVelocity = 0.0;  %  [m/sec]
                WOBAutoDriller = true;
                DesiredWOB = 6000;  %  [kg]
                WOBProportionalGain = 0.5e-5;
                WOBIntegralGain = 0.5e-6;      
            elseif timeStep < 1400
                FlowRateIn = 2200/60000;  %  [m3/sec]
                TopOfStringVelocity = 0.0;  %  [m/sec]
                WOBAutoDriller = true;
                DesiredWOB = 5000;  %  [kg]
                WOBProportionalGain = 0.5e-5;
                WOBIntegralGain = 0.5e-6; 
            else
                FlowRateIn = 2200/60000;  %  [m3/sec]
                TopOfStringVelocity = 0.0;  %  [m/sec]
                WOBAutoDriller = true;
                DesiredWOB = 4000;  %  [kg]
                WOBProportionalGain = 1e-5;
                WOBIntegralGain = 1e-6;
            end
            
            Sim.FlowRateIn = FlowRateIn;
            Sim.TopOfStringVelocity = TopOfStringVelocity;
            Sim.ROP = ROP;
            Sim.WOBAutoDriller = WOBAutoDriller;
            Sim.DesiredWOB = DesiredWOB;
            Sim.WOBProportionalGain = WOBProportionalGain;
            Sim.WOBIntegralGain = WOBIntegralGain;
            Sim.SurfaceRPM = SurfaceRPM;
            Sim.ChokeOpening = ChokeOpening;
            Sim.ChokePumpFlowRateIn = ChokePumpFlowRateIn;
            Sim.DensityIn = DensityIn;
            Sim.BopChokeOpening = BopChokeOpening;          
            
            % Step simulator
            Sim.Step();            

            if ~Sim.IsOK % Simulator fails
                break;
            end
            
            if plotFigures
                % Update animation
                addpoints(ChokePumpFlowRateInLine,timeStep,Sim.ChokePumpFlowRateIn*6E4);
                addpoints(FlowRateInLine,timeStep,Sim.FlowRateIn*6E4);
                addpoints(FlowRateOutLine,timeStep,Sim.FlowRateOut*6E4);
                addpoints(BHPLine,timeStep,Sim.BottomHolePressure/1E5);
                addpoints(SPPLine,timeStep,Sim.SPP/1E5);
                addpoints(ChokePLine,timeStep,Sim.ChokePressure/1E5);
                addpoints(ChokeOpeningActLine,timeStep,Sim.ChokeOpeningStatus*100);
                addpoints(ChokeOpeningCmdLine,timeStep,Sim.ChokeOpening*100);
                [AnnulusPressureResultMD, AnnulusPressureResultVal] = Sim.AnnulusPressureProfile;
                set(AnnulusPressurePlot,'xdata',AnnulusPressureResultVal/1E5, 'ydata', -1*AnnulusPressureResultMD)
                [AnnulusTemperatureResultMD, AnnulusTemperatureResultVal] = Sim.AnnulusTemperatureProfile;
                set(AnnulusTemperaturePlot,'xdata',AnnulusTemperatureResultVal, 'ydata', -1*AnnulusTemperatureResultMD)
            end
            display(['Total step duration: ' num2str(toc(tStartStep))]);
        end
        
        % Stop simulation process
        Sim.Stop;
        
    catch exception
        % Stop simulation
        Sim.Stop;        
        rethrow(exception);
    end
    
end
