% NOTE: You need Matlab R2016b version to run the OpenLab simulator

clear variables;
clc;
close all;

% Add to the matlab path the folder where this file is located and all the
% subfolders
addpath(genpath(pwd));

IdentityServerURL = 'https://live.openlab.app/';

[username, api_key, license_guid] = GetLoginData();

% Select the configuration you want to run simulations on
ConfigurationName = 'Your configuration name';

% Choose a name on your simulation and set some initial parameters
SimulationName = 'Flow sweep with bhp control (Matlab)';
InitialBitDepth = 2500; % [m]
InitialTopOfStringPosition = 20;

% Create simulation object
Sim = OpenLabClient(IdentityServerURL, username,api_key,license_guid,ConfigurationName,SimulationName,InitialBitDepth, InitialTopOfStringPosition);

% Ramp settings
RampIndex = 1;
RampValuesDown = (2500:-100:1500)/60000; % [m3/sec]
RampValuesUp = (1500:100:2500)/60000;  % [m3/sec]
RampValues = [RampValuesDown RampValuesUp]; % [m3/sec]
RampStepDuration = 30; % [sec]
RampStartTime = 100; % [sec]
RampTimeSteps = RampStartTime:RampStepDuration:length(RampValues)*RampStepDuration-1+RampStartTime;

% Time steps to simulate
MaxTimeSteps = (RampTimeSteps(end)+100);

% Controller settings
Kp = -0.015; Ki = Kp/10; Ts = 1;
ReferenceBHPPressure = 380 * 1E5; % [Pa]
InitialChokeOpening = 0.25; % [closed: 0, open: 1]
PI = PIcontroller(Kp, Ki, Ts, ReferenceBHPPressure,InitialChokeOpening); % Create PI controller object, pressure in [Pa]

if Sim.IsOK
    try
        for timeStep = 1 : MaxTimeSteps
            tStartStep = tic;
            
            if timeStep >= RampStartTime % Flow sweep and PI control of the choke
                if (RampIndex < length(RampTimeSteps) && RampIndex < length(RampValues))
                    if (timeStep >= RampTimeSteps(RampIndex) &&  timeStep < RampTimeSteps(RampIndex + 1))
                        FlowRateIn = RampValues(RampIndex);
                        if (timeStep == RampTimeSteps(RampIndex + 1) - 1)
                            RampIndex = RampIndex + 1;
                        end
                    end
                elseif (RampIndex == length(RampTimeSteps) && RampIndex == length(RampValues))
                    FlowRateIn = RampValues(RampIndex);
                end
                
                if timeStep == RampStartTime
                    % Reset PI controller before usage, set reference value and initial output (= initial
                    % choke opening)
                    PI.Reset(ReferenceBHPPressure,Sim.ChokeOpeningStatus);
                end
                
                ChokeOpening  = PI.GetOutput(Sim.BottomHolePressure); % Get choke opening from PI controller
                
            else % Constant flow rate and choke opening
                FlowRateIn = RampValues(1);
                ChokeOpening = InitialChokeOpening;
            end
            
            % Set setpoints to the simulator
            Sim.FlowRateIn = FlowRateIn;  %  [m3/sec]
            Sim.ChokeOpening = ChokeOpening;  %   [0-1]
            Sim.ChokePumpFlowRateIn = 0/60000;  %  [m3/sec]
            Sim.TopOfStringVelocity = 0;  %  [m/sec]
            Sim.SurfaceRPM = 0/60;  %  [revolutions per sec.]
            Sim.ROP = 0/3600;  %  [m/sec]
            
            % Step simulator
            Sim.Step();
            
            if ~Sim.IsOK % Simulator fails
                break;
            end

            display(['Total step duration: ' num2str(toc(tStartStep))]);
        end
        
        % Stop simulation process
        Sim.Stop;
        
    catch exception
        % Stop simulation
        Sim.Stop;        
        rethrow(exception);
    end
    
end
