% NOTE: You need Matlab R2016b version to run the OpenLab simulator

clear variables;
clc;
close all;

% Add to the matlab path the folder where this file is located and all the
% subfolders
addpath(genpath(pwd));

IdentityServerURL = 'https://live.openlab.app/';

[username, api_key, license_guid] = GetLoginData();

ConfigurationName = 'test OWD';
SimulationName = 'Flow sweep with bhp control (Matlab)';
InitialBitDepth = 4997; % [m]
InitialTopOfStringPosition = 15;

% Create simulation object
Sim = OpenLabClient(IdentityServerURL, username,api_key,license_guid,ConfigurationName,SimulationName,InitialBitDepth, InitialTopOfStringPosition);

% Ramp settings
RampIndex = 1;
RampValuesDown = (2500:-100:1500)/60000; % [m3/sec]
RampValuesUp = (1500:100:2500)/60000;  % [m3/sec]
RampValues = [RampValuesDown RampValuesUp]; % [m3/sec]
RampStepDuration = 60; % [sec]
RampStartTime = 130; % [sec]
RampTimeSteps = RampStartTime:RampStepDuration:length(RampValues)*RampStepDuration-1+RampStartTime;

% Time steps to simulate
MaxTimeSteps = (RampTimeSteps(end)+100);

% Controller settings
Kp = 2E-9; Ts = 1;
Ti = 7;
ReferenceBHPPressure = 990 * 1E5; % [Pa]
InitialLiftPumpFlowRate = 0; 
% PI = PIcontroller(Kp, Ki, Ts, ReferenceBHPPressure,InitialChokeOpening); % Create PI controller object, pressure in [Pa]
ReferenceMaxRateOfChange = 1E5; 
IsReversed = true;
MinOutput = -7000/60000;
MaxOutput = 7000/60000;
PI = PIcontrollerV2(Kp, Ti, Ts, ReferenceBHPPressure, InitialLiftPumpFlowRate, ReferenceMaxRateOfChange,IsReversed, MinOutput, MaxOutput);

if Sim.IsOK
    try
        for timeStep = 1 : MaxTimeSteps
            tStartStep = tic;

             Sim.UseContinuousCirculationSystem = true;

            if timeStep > 1000
                ReferenceBHPPressure = 1000E5;
            end
            
            if timeStep >= RampStartTime % Flow sweep and PI control of the lift pump
                if (RampIndex < length(RampTimeSteps) && RampIndex < length(RampValues))
                    if (timeStep >= RampTimeSteps(RampIndex) &&  timeStep < RampTimeSteps(RampIndex + 1))
                        FlowRateIn = RampValues(RampIndex);
                        if (timeStep == RampTimeSteps(RampIndex + 1) - 1)
                            RampIndex = RampIndex + 1;
                        end
                    end
                elseif (RampIndex == length(RampTimeSteps) && RampIndex == length(RampValues))
                    FlowRateIn = RampValues(RampIndex);
                end
                
                if timeStep == RampStartTime
                    % Reset PI controller before usage, set reference value and initial output (= initial
                    % lift pump flow rate)
                    PI.Reset(ReferenceBHPPressure,Sim.CalculatedLiftPumpFlowRate);
                end                
                
                LiftPumpFlowRate  = PI.GetOutput(ReferenceBHPPressure,Sim.BottomHolePressure); % Get lift pump flow rate from PI controller
                LiftPumpControlMode = 1; % Send flow rate to lift pump
                
            else
                LiftPumpControlMode = 0; % Need to start with a bypass mode
                FlowRateIn = RampValues(1);
                LiftPumpFlowRate = InitialLiftPumpFlowRate;
            end
            
            % Set setpoints to the simulator
            Sim.FlowRateIn = FlowRateIn;  %  [m3/sec]
            Sim.TopOfStringVelocity = 0;  %  [m/sec]
            Sim.SurfaceRPM = 0/60;  %  [revolutions per sec.]
            Sim.ROP = 0/3600;  %  [m/sec]
            Sim.SubsurfaceValve = 1; % subsurface valve
            Sim.LiftPumpFlowRate = LiftPumpFlowRate;
            Sim.LiftPumpControlMode = LiftPumpControlMode;
            
            % Step simulator
            Sim.Step();
            
            if ~Sim.IsOK % Simulator fails
                break;
            end

            display(['Total step duration: ' num2str(toc(tStartStep))]);
        end
        
        % Stop simulation process
        Sim.Stop;
        
    catch exception
        % Stop simulation
        Sim.Stop;        
        rethrow(exception);
    end
    
end
