classdef OpenLabClient < handle
    
    %   Constructor:
    %       sim = OpenLabClient(username, api key, configuration name, simulation name, initial bit depth) creates an instance of
    %       OpenLabClient,step
    %             initial bit depth is optional - if not specified then the initial bit depth is set to TD
    %       e.g. sim = OpenLabClient('test.testesen@iris.no','7D987A0CB45697F3CB23D3B928A1622E6C73133EEFDF5DB6A81D19DF91E2EB3B','TestWell-Drilling.xml','Sonjas Matlab sim', 2400)
    %
    %   Methods:
    %       Set setpoints:
    %           TopOfStringVelocity
    %           FlowRateIn
    %           SurfaceRPM
    %           ChokePumpFlowRateIn
    %           ChokeOpening
    %           BopChokeOpening
    %           MainPitMainPump (optional, bool)
    %           MainPitChokePump (optional, bool)
    %           MainPitKillPump (optional, bool)
    %           MainPitReturn (optional, bool)
    %           ControlMainPit (optional, bool)
    %           ControlMainPitDensity (optional, double, used if ControlMainPit = true)
    %           ControlMainPitTemperature (optional, double, used if ControlMainPit = true)
    %           DensityIn (optional, double, This will override ControlMainPit(set to true) and ControlMainPitDensity (set to DensityIn)
    %           WOBAutoDriller
    %           ROP
    %           DesiredWOB
    %           WOBProportionalGain
    %           WOBIntegralGain
    %           LiftPumpControlMode
    %           RigChokeOpening
    %           KillPumpFlowRateIn
    %           ChokelineValve
    %           KillineValve
    %           LiftPumpFlowRate
    %           SubsurfaceValve
    %           UseContinuousCirculationSystem

    
    %
    %       Step simulator:
    %           Step
    %
    %       Get results:
    %           FlowRateOut
    %           HookLoad
    %           SPP
    %           BitPressure
    %           BottomHolePressure
    %           ChokePressure
    %           BopChokePressure
    %           DownholeECD
    %           BitDepth
    %           TD
    %           GasFlowRateOut
    %           TotalInfluxMass
    %           TotalMudLossMass
    %           ChokeOpening
    %           BopChokeOpening
    %           WOB
    %           ActualROP
    %           SurfaceTorque
    %           FluidTemperatureOut
    %           FluidTemperatureIn
    %           TopOfStringVelocityActual
    %           ChokePumpFlowRateInActual
    %           FlowRateInActual
    %           SurfaceRPMActual
    %           CalculatedLiftPumpFlowRate
    %           RigChokeOpeningActual
    %           RigChokePressure
    %           KillPumpFlowRateInActual
    
    
    properties (Access = public)
        % setpoints
        TopOfStringVelocity = 0;
        FlowRateIn = 0;
        SurfaceRPM = 0;
        ChokePumpFlowRateIn = 0;
        ChokeOpening = 1;
        BopChokeOpening = 1;
        MainPitMainPump = nan;
        MainPitChokePump = nan;
        MainPitKillPump = nan;
        MainPitReturn = nan;
        ControlMainPit = nan;
        ControlMainPitDensity = nan;
        DensityIn = nan;
        ControlMainPitTemperature = nan;
        
                    
        LiftPumpControlMode = nan;
        RigChokeOpening = nan;
        KillPumpFlowRateIn = nan;
        ChokelineValve = nan;
        KillineValve = nan;
        LiftPumpFlowRate = nan;
        SubsurfaceValve = nan;
        
        ROP = 0;
        WOBAutoDriller = false;
        DesiredWOB = 0;
        WOBProportionalGain = 0.000001;
        WOBIntegralGain = 1e-8;
        UseContinuousCirculationSystem = false;
        
    end
    
    properties(Access = private, Constant) % can not be changed by the user
        TimeOutInitialization = 60; % sec
        TimeOutTryStepAgain = 60; % sec
        TimeOutGetResults = 60; % sec
        
        CheckIfSimulatorIsInitializedInterval = 0.5; % sec
        CheckForNewResultsInterval = 0.025; % sec
    end
    
    properties (Access = private) % can not be set by the user
        WebServerURL;
        IdentityServerURL;
        Headers;
        Tokens;
        ConfigID;
        SimID;
        TimeStep = 0;
        MaxTimeStep;
        Initialized = false;
        OK = true;
        SetPointBody;
        DepthSubscription;
        SubscriptionString;
        SimulateInfluxLoss;
        
        % results
        FlowRateOutResult = 0;
        FlowRateInResult = 0;
        RigChokeOpeningResult = 0;
        RigChokePressureResult = 0;
        KillPumpFlowRateInResult = 0;
        ChokePumpFlowRateInResult = 0;
        HookLoadResult = 0;
        SurfaceRPMResult = 0;
        FluidTemperatureOutResult = 0;
        FluidTemperatureInResult = 0;
        SPPResult = 0;
        BitPressureResult = 0;
        BottomHolePressureResult = 0;
        LiftPumpFlowRateResult = 0;
        ChokePressureResult = 0;
        DownholeECDResult = 0;
        BopChokePressureResult = 0
        BitDepthResult = 0;
        TopOfStringVelocityResult = 0;
        TDResult = 0;
        GasFlowRateOutResult = 0;
        TotalInfluxMassResult = 0;
        TotalMudLossMassResult = 0;
        ChokeOpeningResult = 0;
        BopChokeOpeningResult = 0;
        WOBResult = 0;
        ROPResult = 0;
        SurfaceTorqueResult = 0;
        
        AnnulusPressureResultMD = [];
        AnnulusPressureResultVal = [];
        AnnulusTemperatureResultMD = [];
        AnnulusTemperatureResultVal = [];
        GasVolumeFractionResultMD = [];
        GasVolumeFractionResultVal = [];
        ECDResultMD = [];
        ECDResultVal = [];
        
        SimulationStartTime;
        TimeAtStep10 = tic;
    end
    
    methods
        %% Login, create simulation and initialize
        function this = OpenLabClient(IdentityServerURL, username, api_key,license_guid, ConfigurationName,SimulationName,InitialBitDepth,TopOfStringPosition,...
                UseReservoirModel,ManualReservoirMode,ManualInfluxLossMassRate,ManualInfluxLossTotalMass,ManualInfluxLossMD,ComplexReservoirKickOffTime,...
                UseTransientMechanicalModel,StepDuration)
            this.IdentityServerURL = IdentityServerURL;
            
            this.DepthSubscription = true;
            
            if ~exist('UseReservoirModel','var') || isempty(UseReservoirModel); UseReservoirModel = false; end
            if ~exist('ManualReservoirMode','var') || isempty(ManualReservoirMode); ManualReservoirMode = false; end
            if ~exist('ManualInfluxLossMassRate','var') || isempty(ManualInfluxLossMassRate); ManualInfluxLossMassRate = 0; end
            if ~exist('ManualInfluxLossTotalMass','var') || isempty(ManualInfluxLossTotalMass); ManualInfluxLossTotalMass = 0; end
            if (~exist('ManualInfluxLossMD','var') || isempty(ManualInfluxLossMD)) && (~exist('BitDepth','var') || isempty(BitDepth)); ManualInfluxLossMD = 0;
            elseif (~exist('ManualInfluxLossMD','var') || isempty(ManualInfluxLossMD)); ManualInfluxLossMD = InitialBitDepth; end
            if ~exist('ComplexReservoirKickOffTime','var') || isempty(ComplexReservoirKickOffTime); ComplexReservoirKickOffTime = 60; end
            if ~exist('UseTransientMechanicalModel','var') || isempty(UseTransientMechanicalModel); UseTransientMechanicalModel = false; end
            if ~exist('StepDuration','var') || isempty(StepDuration); StepDuration = 1; end
            
            if UseReservoirModel && ~ManualReservoirMode
                Id = 'GeoPressureGradient';
            elseif ~UseReservoirModel && ~ManualReservoirMode
                Id = 'NoInfluxLoss';
            elseif UseReservoirModel && ManualReservoirMode && ManualInfluxLossMassRate >= 0
                Id = 'ManualInflux';
            else
                Id = 'ManualLoss';
            end
            
            this.SimulateInfluxLoss = UseReservoirModel;
            
            %% Login
            url = [this.IdentityServerURL 'connect/token' ];
            
            encodedbody = strcat('client_id=OpenLab&grant_type=openlab_api_key&username=', username, '&key=', api_key);
            
            try
                [output,extras]  = urlread2(url,'post',encodedbody);
            catch ex
                throw(addCause(MException('OpenLab:authenticationFailed', 'Failed to authenticate against OpenLab.'), ex));
            end
            
            if ~extras.isGood
                throw(MException('OpenLab:authenticationFailed', strjoin(cat(1, {'Received invalid authentication response:'}, struct2cell(jsondecode(output))), '\n')));
            end
            
            this.Tokens = jsondecode(output);
            
            %% Retrieve WebServerURL
            
            contentType = 'application/json; version=1';
            
            whoami_url = [this.IdentityServerURL 'users/whoami'];
            whoami_headers = struct('name', {'Authorization', 'Referer', 'Content-Type', 'Accept'}, 'value', {sprintf('Bearer %s', this.Tokens.access_token), this.IdentityServerURL contentType contentType});
            
            try
                [output, extras] = urlread2(whoami_url,'get', '', whoami_headers);
            catch ex
                throw(addCause(MException('OpenLab:userInfoFailed', 'Failed to get user information from OpenLab.'), ex));
            end
            
            if ~extras.isGood
                throw(MException('OpenLab:userInfoFailed', strjoin(cat(1, {'Failed to get user information from OpenLab:'}, struct2cell(jsondecode(output))), '\n')));
            end
            
            userData = jsondecode(output);
            
            try
                this.WebServerURL = strcat(getfield(userData, 'EndpointUrl'), '/');
            catch
                this.WebServerURL = this.IdentityServerURL;
            end
            
            this.Headers = struct('name', {'Authorization', 'Referer', 'Content-Type', 'Accept'}, 'value', {sprintf('Bearer %s', this.Tokens.access_token), this.WebServerURL contentType contentType});
            
            %% Get configuration
            if exist('ConfigurationName','var') && ~isempty(ConfigurationName)
                url_ = [this.WebServerURL 'configurations/' ConfigurationName ];
                
                url = strrep(url_,' ','%20');
                [output,extras] = this.readapi(url,'get', []);
                
                configInfo = jsondecode(output);
                if isfield(configInfo,'Errors')
                    error([configInfo.Title configInfo.Detail ]);
                end
                
                this.OK = extras.isGood;
                this.ConfigID = char(output(2:end-1));
                
                if ~extras.isGood
                    error('Can not find the specified configuration in the OpenLab Database. Check if the configuration name is correct. %s %s', extras.status.msg, output);
                end
                if strcmp(this.ConfigID,'00000000-0000-0000-0000-000000000000')
                    this.OK = false;
                    error('The specified configuration does not exist in the OpenLab Database.');
                end
                
                %% Create simulation
                url = [this.WebServerURL 'simulations' ];
                
                this.SimID = char(java.util.UUID.randomUUID());
                clear body;
                body.ConfigurationID = this.ConfigID;
                body.SimulationID = this.SimID;
                body.Name = SimulationName;
                body.LicenseId = license_guid;
                
                if exist('InitialBitDepth','var') && ~isempty(InitialBitDepth)
                    body.InitialValues.BitDepth = InitialBitDepth;
                end
                
                if exist('TopOfStringPosition','var') && ~isempty(TopOfStringPosition)
                    body.InitialValues.TopOfStringPosition = TopOfStringPosition;
                end
                
                body.ModelConfiguration.UseReservoirModel = UseReservoirModel;
                body.ModelConfiguration.ManualReservoirMode = ManualReservoirMode;
                body.ModelConfiguration.ManualInfluxLossMassRate = ManualInfluxLossMassRate;
                body.ModelConfiguration.ManualInfluxLossTotalMass = ManualInfluxLossTotalMass;
                body.ModelConfiguration.ManualInfluxLossMD = ManualInfluxLossMD;
                body.ModelConfiguration.ComplexReservoirKickOffTime = ComplexReservoirKickOffTime;
                body.ModelConfiguration.UseTransientMechanicalModel = UseTransientMechanicalModel;
                body.ModelConfiguration.StepDuration = StepDuration;
                body.ModelConfiguration.Id = Id;
                
                [output,extras] = this.readapi(url,'post', jsonencode(body));
                
                this.OK = extras.isGood;
                
                if ~this.OK
                    if extras.status.value == 429
                        throw(MException('OpenLab:createSimulationFailed', 'Maximum number of concurrent simulations reached. Please stop some simulations before creating new ones.'));
                    elseif extras.status.value == 422
                        throw(MException('OpenLab:createSimulationFailed', strjoin(cellstr(getfield(jsondecode(output), 'Errors')), '\n')));
                    else
                        throw(MException('OpenLab:createSimulationFailed', strjoin(cat(1, {'Failed to create OpenLab simulation:'}, struct2cell(jsondecode(output))), '\n')));
                    end
                end
                
                simInfo = jsondecode(output);
                
                this.SimID = simInfo.SimulationID;
                
                pause(2);
                
                %% Wait for simulation to start
                url = [this.WebServerURL 'simulations/'  this.SimID  '/status' ];
                
                status = '';
                tries = 0;
                disp('Initializing the OpenLab Simulation...');
                while ~strcmp(status,'Running')
                    [output,extras] = this.readapi(url,'get',[]);
                    output = jsondecode(output);
                    if isfield(output,'Errors')
                        error([output.Title output.Detail]);
                    end
                    status = output.Status;
                    this.OK = extras.isGood;
                    if ~this.OK
                        error('Failed to initialize the OpenLab Simulation.');
                    end
                    if strcmp(status,'Running')
                        this.Initialized = true;
                        break;
                    else
                        pause(this.CheckIfSimulatorIsInitializedInterval);
                    end
                    
                    if tries>this.TimeOutInitialization/this.CheckIfSimulatorIsInitializedInterval
                        this.OK = false;
                        error('Failed to initialize the OpenLab Simulation.')
                    end
                    tries = tries + 1;
                end
                disp('OpenLab Simulation is initialized...')
                
                pause(3);
                
                %% Create ResultQuery
               
                if UseReservoirModel == false; disp('No influx/loss'); end
                if UseReservoirModel == true && ManualReservoirMode == false; disp('Influx/loss based on geopressures'); end
                if UseReservoirModel == true && ManualReservoirMode == true; disp('Manual influx/loss'); end
                
                
                if this.SimulateInfluxLoss
                    this.SubscriptionString = [ ...
                        '&SPP=true' ...
                        '&DownholePressure=true' ...
                        '&CalculatedPressureBottomHole=true' ...
                        '&FlowRateOut=true' ...
                        '&FlowRateIn=true' ...
                        '&ChokePumpFlowRateIn=true' ...
                        '&HookLoad=true' ...
                        '&SurfaceRPM=true' ...
                        '&TopOfStringVelocity=true' ...
                        '&FluidTemperatureOut=true' ...
                        '&FluidTemperatureIn=true' ...
                        '&SurfaceTorque=true' ...
                        '&BitDepth=true' ...
                        '&TD=true' ...
                        '&ChokePressure=true' ...
                        '&DownholeECD=true' ...
                        '&BopChokePressure=true' ...
                        '&BitDepth=true' ...
                        '&ChokeOpening=true' ...
                        '&BopChokeOpening=true' ...
                        '&TotalInfluxMass=true' ...
                        '&GasFlowRateOut=true' ...
                        '&TotalMudLossMass=true' ...
                        '&AnnulusPressure=true' ...
                        '&AnnulusTemperature=true' ...
                        '&GasVolumeFraction=true' ...
                        '&WOB=true' ...
                        '&InstantaneousROP=true' ...
                        '&CalculatedLiftPumpFlowRate=true' ...
                        ];
                else
                    this.SubscriptionString = [ ...
                        '&SPP=true' ...
                        '&DownholePressure=true' ...
                        '&CalculatedPressureBottomHole=true' ...
                        '&FlowRateOut=true' ...
                        '&FlowRateIn=true' ...
                        '&ChokePumpFlowRateIn=true' ...
                        '&HookLoad=true' ...
                        '&SurfaceRPM=true' ...
                        '&TopOfStringVelocity=true' ...
                        '&FluidTemperatureOut=true' ...
                        '&FluidTemperatureIn=true' ...
                        '&SurfaceTorque=true' ...
                        '&BitDepth=true' ...
                        '&TD=true'...
                        '&ChokePressure=true' ...
                        '&DownholeECD=true' ...
                        '&BopChokePressure=true' ...
                        '&BitDepth=true&TD=true' ...
                        '&ChokeOpening=true' ...
                        '&BopChokeOpening=true' ...
                        '&AnnulusPressure=true' ...
                        '&AnnulusTemperature=true' ...
                        '&WOB=true' ...
                        '&InstantaneousROP=true' ...
                        '&CalculatedLiftPumpFlowRate=true' ...
                        ];
                end
                
                this.SimulationStartTime = tic;
            end
        end
        
        function [output, extras] = readapi(this, url, method, body)
            [output, extras] = urlread2(url, method, body, this.Headers);
            if ~extras.isGood && (extras.status.value == 401)
                %% Refresh
                refresh_url = [this.IdentityServerURL 'connect/token' ];
                refresh_body = strcat('client_id=OpenLab&grant_type=refresh_token&refresh_token=', this.Tokens.refresh_token);
                refresh_header = struct('name', {'Content-Type'}, 'value', {'application/x-www-form-urlencoded;charset=UTF-8'});
                try
                    [response,status]  = urlread2(refresh_url,'post',refresh_body,refresh_header);
                catch ex
                    throw(addCause(MException('OpenLab:tokenRefreshFailed', 'Failed to re-authenticate using refresh token.'), ex));
                end
                this.OK = status.isGood;
                if ~status.isGood
                    throw(MException('OpenLab:tokenRefreshFailed', strjoin(cat(1, {'Failed to re-authenticate using refresh token:'}, struct2cell(jsondecode(output))), '\n')));
                end
                
                this.Tokens = jsondecode(response);
                contentType = 'application/json; version=1';
                
                this.Headers = struct('name', {'Authorization', 'Referer', 'Content-Type', 'Accept'}, 'value', {sprintf('Bearer %s', this.Tokens.access_token), this.WebServerURL contentType contentType});
                [output, extras] = urlread2(url, method, body, this.Headers);
            end
        end
        
        %% Step simulator
        function [] = Step(this)
            if this.Initialized == false
                error('Trying to step, but the OpenLab Simulation is not initialized. Restart the program.')
            end
            
            this.TimeStep = this.TimeStep + 1;
            if this.TimeStep == 10
                this.TimeAtStep10 = tic;
            end
            
            display(['Step no: ' num2str(this.TimeStep)])
            %             tStartStep = tic;
            
            %% Set setpoints - step simulator
            url = [this.WebServerURL 'simulations/'  this.SimID  '/setpoints' ];
            clear this.SetPointBody
            this.SetPointBody.TimeStep = this.TimeStep;
            this.SetPointBody.Data.TopOfStringVelocity = this.TopOfStringVelocity;
            this.SetPointBody.Data.FlowRateIn = this.FlowRateIn;
            this.SetPointBody.Data.SurfaceRPM = this.SurfaceRPM;
            this.SetPointBody.Data.ChokePumpFlowRateIn = this.ChokePumpFlowRateIn;
            this.SetPointBody.Data.ChokeOpening = this.ChokeOpening;
            this.SetPointBody.Data.BopChokeOpening = this.BopChokeOpening;
            
            if ~isnan(this.MainPitMainPump)
                this.SetPointBody.Data.MainPitMainPump = this.MainPitMainPump;
            end
            if ~isnan(this.MainPitKillPump)
                this.SetPointBody.Data.MainPitKillPump = this.MainPitKillPump;
            end
            if ~isnan(this.MainPitChokePump)
                this.SetPointBody.Data.MainPitChokePump = this.MainPitChokePump;
            end
            if ~isnan(this.MainPitReturn)
                this.SetPointBody.Data.MainPitReturn = this.MainPitReturn;
            end
            if ~isnan(this.ControlMainPit)
                this.SetPointBody.Data.ControlMainPit = this.ControlMainPit;
            end
            if ~isnan(this.ControlMainPitDensity)
                this.SetPointBody.Data.ControlMainPitDensity = this.ControlMainPitDensity;
            end
            if ~isnan(this.ControlMainPitTemperature)
                this.SetPointBody.Data.ControlMainPitTemperature = this.ControlMainPitTemperature;
            end
            if ~isnan(this.DensityIn)
                this.SetPointBody.Data.DensityIn = this.DensityIn;
            end
            
            if ~isnan(this.DensityIn)
                this.SetPointBody.Data.DensityIn = this.DensityIn;
            end
            
            if ~isnan(this.LiftPumpControlMode)
                this.SetPointBody.Data.LiftPumpControlMode = this.LiftPumpControlMode;
            end
            if ~isnan(this.RigChokeOpening)
                this.SetPointBody.Data.RigChokeOpening = this.RigChokeOpening;
            end
            if ~isnan(this.KillPumpFlowRateIn)
                this.SetPointBody.Data.KillPumpFlowRateIn = this.KillPumpFlowRateIn;
            end
            if ~isnan(this.ChokelineValve)
                this.SetPointBody.Data.ChokelineValve = this.ChokelineValve;
            end
            if ~isnan(this.KillineValve)
                this.SetPointBody.Data.KillineValve = this.KillineValve;
            end
            if ~isnan(this.LiftPumpFlowRate)
                this.SetPointBody.Data.LiftPumpFlowRate = this.LiftPumpFlowRate;
            end
            if ~isnan(this.SubsurfaceValve)
                this.SetPointBody.Data.SubsurfaceValve = this.SubsurfaceValve;
            end
            
            
            this.SetPointBody.Data.WOBAutoDriller = this.WOBAutoDriller;
            this.SetPointBody.Data.DesiredWOB = this.DesiredWOB;
            this.SetPointBody.Data.WOBProportionalGain =this.WOBProportionalGain;
            this.SetPointBody.Data.WOBIntegralGain = this.WOBIntegralGain;

            this.SetPointBody.Data.UseContinuousCirculationSystem = this.UseContinuousCirculationSystem;
            
            this.SetPointBody.Data.DesiredROP = this.ROP;
            
            [output,extras] = this.readapi(url,'post', jsonencode({this.SetPointBody}));
            
            this.OK = extras.isGood;
            
            if ~extras.isGood
                if extras.status.value == 409
                    warning('Setpoint for step %d already exists.', this.TimeStep);
                    this.OK = true;
                    return;
                elseif extras.status.value == 422
                    throw(MException('OpenLab:stepFailed', strjoin(cellstr(getfield(jsondecode(output), 'Detail')), '\n')));
                else
                    throw(MException('OpenLab:stepFailed', strjoin(cat(1, {'Failed to send setpoint to OpenLab:'}, struct2cell(jsondecode(output))), '\n')));
                end
            end
            
            pause(0.01);
            
            %% Get simulation results
            output = '[]';
            tries = 0;
            extras = '[]';
            
            ResultUrl = [this.WebServerURL 'simulations/'  this.SimID  '/results/?timeStepFrom=' ...
                num2str(this.TimeStep) '&timeStepTo=' num2str(this.TimeStep) this.SubscriptionString];
            
            while strcmp(output,'[]') || strcmp(extras,'[]') || ~extras.isGood
                [output,extras] = this.readapi(ResultUrl,'get', []);
                this.OK = extras.isGood;
                if strcmp(output,'[]') || ~extras.isGood
                    pause(this.CheckForNewResultsInterval);
                end
                if isfield(output,'Errors')
                    error([output.Title output.Detail]);
                end
                pause(0.01);
                tries = tries + 1;
                
                if ( strcmp(output,'[]') || strcmp(extras,'[]') )
                    % Check simulation status
                    url = [this.WebServerURL 'simulations/'  this.SimID  '/status' ];
                    [statusOutput,extras] = this.readapi(url,'get',[]);
                    if isfield(statusOutput,'Errors')
                        error([statusOutput.Title statusOutput.Detail]);
                    end
                    this.OK = extras.isGood;
                    statusOutput = jsondecode(statusOutput);
                    status = statusOutput.Status;
                    if isfield(statusOutput,'Description')
                        description = statusOutput.Description;
                    else
                        description = '';
                    end
                    
                    if strcmp(status,'Completed')
                        this.OK = false;
                        error(['OpenLab Simulation is completed. ' description '. Update your account.']);
                    elseif strcmp(status,'Aborted')
                        this.OK = false;
                        error('OpenLab Simulation is aborted.');
                    end
                end
                
                if tries>this.TimeOutGetResults/this.CheckForNewResultsInterval
                    disp('Failed to receive results from the OpenLab Simulation.')
                    this.OK = false;
                    return;
                end
            end
            
            %% Parse simulation results
            %             tStartParse = tic;
            if ~(strcmp(output,'[]')) && ~strcmp(extras,'[]') && extras.isGood
                try
                    %                     t2 = tic;
                    results = jsondecode(output);
                    if isfield(results,'Errors')
                        error([results.Title results.Detail]);
                    end
                    %                     display(['jsondecode: ' num2str(toc(t2))]);
                    if ~iscell(results)
                        results = num2cell(results);
                    end
                    
                    this.FlowRateOutResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'FlowRateOut'}),1)}.v.v;
                    this.FlowRateInResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'FlowRateIn'}),1)}.v.v;
                    this.ChokePumpFlowRateInResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'ChokePumpFlowRateIn'}),1)}.v.v;
                    this.HookLoadResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'HookLoad'}),1)}.v.v;
                    this.SurfaceRPMResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'SurfaceRPM'}),1)}.v.v;
                    this.TopOfStringVelocityResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'TopOfStringVelocity'}),1)}.v.v;
                    this.FluidTemperatureOutResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'FluidTemperatureOut'}),1)}.v.v;
                    this.FluidTemperatureInResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'FluidTemperatureIn'}),1)}.v.v;
                    this.SurfaceTorqueResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'SurfaceTorque'}),1)}.v.v;
                    this.SPPResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'SPP'}),1)}.v.v;
                    this.BitPressureResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'DownholePressure'}),1)}.v.v;
                    this.BottomHolePressureResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'CalculatedPressureBottomHole'}),1)}.v.v;
                    
                    try 
                         this.LiftPumpFlowRateResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'CalculatedLiftPumpFlowRate'}),1)}.v.v;
                    catch
                         this.LiftPumpFlowRateResult = 0;
                    end
                    try 
                         this.RigChokeOpeningResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'RigChokeOpening'}),1)}.v.v;
                    catch
                         this.RigChokeOpeningResult = 0;
                    end
                    try 
                         this.KillPumpFlowRateInResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'KillPumpFlowRateIn'}),1)}.v.v;
                    catch
                         this.KillPumpFlowRateInResult = 0;
                    end
                    try 
                         this.RigChokePressureResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'RigChokePressure'}),1)}.v.v;
                    catch
                         this.RigChokePressureResult = 0;
                    end
                    this.ChokePressureResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'ChokePressure'}),1)}.v.v;
                    this.DownholeECDResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'DownholeECD'}),1)}.v.v;
                    this.BopChokePressureResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'BopChokePressure'}),1)}.v.v;
                    this.BitDepthResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'BitDepth'}),1)}.v.v;
                    this.TDResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'TD'}),1)}.v.v;
                    this.ChokeOpeningResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'ChokeOpening'}),1)}.v.v;
                    this.BopChokeOpeningResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'BopChokeOpening'}),1)}.v.v;
                    this.ROPResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'InstantaneousROP'}),1)}.v.v;
                    this.WOBResult = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'WOB'}),1)}.v.v;
                    if this.SimulateInfluxLoss
                        temp = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'TotalMudLossMass'}),1)}.v.v;
                        if strcmp(temp,'NaN')
                            this.TotalMudLossMassResult = NaN;
                        else
                            this.TotalMudLossMassResult = temp;
                        end
                        temp = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'TotalInfluxMass'}),1)}.v.v;
                        if strcmp(temp,'NaN')
                            this.TotalInfluxMassResult = NaN;
                        else
                            this.TotalInfluxMassResult = temp;
                        end
                        
                        temp = results{find(strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'GasFlowRateOut'}),1)}.v.v;
                        if strcmp(temp,'NaN')
                            this.GasFlowRateOutResult = NaN;
                        else
                            this.GasFlowRateOutResult = temp;
                        end
                        
                        idx = strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'GasVolumeFraction'});
                        table = struct2table(results{idx}.v);
                        
                        this.GasVolumeFractionResultMD =  table.d;
                        this.GasVolumeFractionResultVal = table.v;
                    end
                    
                    idx = strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'AnnulusPressure'});
                    table = struct2table(results{idx}.v);
                    this.AnnulusPressureResultMD =  table.d; %cell2mat(cellfun( @(item) item.d, results(idx), 'uni', false ));
                    this.AnnulusPressureResultVal = table.v; %cell2mat(cellfun( @(item) item.v, results(idx), 'uni', false ));

                    idx = strcmp( cellfun( @(item) item.t, results, 'uni', false ), {'AnnulusTemperature'});
                    table = struct2table(results{idx}.v);
                    this.AnnulusTemperatureResultMD =  table.d; %cell2mat(cellfun( @(item) item.d, results(idx), 'uni', false ));
                    this.AnnulusTemperatureResultVal = table.v; %cell2mat(cellfun( @(item) item.v, results(idx), 'uni', false ));
                    
                catch
                    disp(output);
                    this.OK = false;
                    return;
                end
            end
            %              display(['Parsing duration: ' num2str(toc(tStartParse))]);
            %             display(['Num tries to get results: ' num2str(tries)])
            %             display(['Total step duration: ' num2str(toc(tStartStep))]);
        end
        
        %% Get simulation properties
        function FlowRateOut = FlowRateOut(this); FlowRateOut = this.FlowRateOutResult;end
        function FlowRateInActual = FlowRateInActual(this); FlowRateInActual = this.FlowRateInResult;end
        function RigChokeOpeningActual = RigChokeOpeningActual(this); RigChokeOpeningActual = this.RigChokeOpeningResult;end
        function KillPumpFlowRateInActual = KillPumpFlowRateInActual(this); KillPumpFlowRateInActual = this.KillPumpFlowRateInResult;end
        function RigChokePressure = RigChokePressure(this); RigChokePressure = this.RigChokePressureResult;end
        function ChokePumpFlowRateInActual = ChokePumpFlowRateInActual(this); ChokePumpFlowRateInActual = this.ChokePumpFlowRateInResult;end
        function HookLoad = HookLoad(this); HookLoad = this.HookLoadResult;end
        function SurfaceRPMActual = SurfaceRPMActual(this); SurfaceRPMActual = this.SurfaceRPMResult;end
        function TopOfStringVelocityActual = TopOfStringVelocityActual(this); TopOfStringVelocityActual = this.TopOfStringVelocityResult;end
        function FluidTemperatureOut = FluidTemperatureOut(this); FluidTemperatureOut = this.FluidTemperatureOutResult;end
        function FluidTemperatureIn = FluidTemperatureIn(this); FluidTemperatureIn = this.FluidTemperatureInResult;end
        function SurfaceTorque = SurfaceTorque(this); SurfaceTorque = this.SurfaceTorqueResult;end
        function SPP = SPP(this); SPP = this.SPPResult;end
        function BitPressure = BitPressure(this); BitPressure = this.BitPressureResult;end
        function BottomHolePressure = BottomHolePressure(this); BottomHolePressure = this.BottomHolePressureResult; end
        function CalculatedLiftPumpFlowRate = CalculatedLiftPumpFlowRate(this); CalculatedLiftPumpFlowRate = this.LiftPumpFlowRateResult; end
        function ChokePressure = ChokePressure(this); ChokePressure = this.ChokePressureResult; end
        function DownholeECD = DownholeECD(this); DownholeECD = this.DownholeECDResult; end
        function BopChokePressure = BopChokePressure(this); BopChokePressure = this.BopChokePressureResult; end
        function GasFlowRateOut = GasFlowRateOut(this); GasFlowRateOut = this.GasFlowRateOutResult; end
        function TotalInfluxMass = TotalInfluxMass(this); TotalInfluxMass = this.TotalInfluxMassResult; end
        function TotalMudLossMass = TotalMudLossMass(this); TotalMudLossMass = this.TotalMudLossMassResult; end
        function ChokeOpeningStatus = ChokeOpeningStatus(this); ChokeOpeningStatus = this.ChokeOpeningResult; end
        function BopChokeOpeningStatus = BopChokeOpeningStatus(this); BopChokeOpeningStatus = this.BopChokeOpeningResult; end
        function BitDepth = BitDepth(this); BitDepth = this.BitDepthResult; end
        function TD = TD(this); TD = this.TDResult; end
        function ActualROP = ActualROP(this); ActualROP = this.ROPResult; end
        function WOB = WOB(this); WOB = this.WOBResult; end
        function SimulationStep = SimulationStep(this); SimulationStep = this.TimeStep; end
        function SimulationID = SimulationID(this); SimulationID = this.SimID; end
        function ConfigurationID = ConfigurationID(this); ConfigurationID = this.ConfigID; end
        function OK = IsOK(this); OK = this.OK; end
        
        function AnnulusPressure = AnnulusPressure(this,depth)
            idx = find(this.AnnulusPressureResultMD >= depth , 1);
            
            if idx == 1
                AnnulusPressure = this.AnnulusPressureResultVal(1);
                return;
            elseif isempty(idx)
                AnnulusPressure = this.AnnulusPressureResultVal(end);
                return;
            else
                depth1 = this.AnnulusPressureResultMD(idx);
                depth2 = this.AnnulusPressureResultMD(idx-1);
                
                val1 = this.AnnulusPressureResultVal(idx);
                val2 = this.AnnulusPressureResultVal(idx-1);
                
                AnnulusPressure=interp1([depth1 depth2],[val1 val2],depth);
                return;
            end
        end

        function AnnulusTemperature = AnnulusTemperature(this,depth)
            idx = find(this.AnnulusTemperatureResultMD >= depth , 1);
            
            if idx == 1
                AnnulusTemperature = this.AnnulusTemperatureResultVal(1);
                return;
            elseif isempty(idx)
                AnnulusTemperature = this.AnnulusTemperatureResultVal(end);
                return;
            else
                depth1 = this.AnnulusTemperatureResultMD(idx);
                depth2 = this.AnnulusTemperatureResultMD(idx-1);
                
                val1 = this.AnnulusTemperatureResultVal(idx);
                val2 = this.AnnulusTemperatureResultVal(idx-1);
                
                AnnulusTemperature=interp1([depth1 depth2],[val1 val2],depth);
                return;
            end
        end
        
        function [GasVolumeFractionResultMD, GasVolumeFractionResultVal ] = GasVolumeFractionProfile(this)
            GasVolumeFractionResultMD = this.GasVolumeFractionResultMD;
            GasVolumeFractionResultVal = this.GasVolumeFractionResultVal;
            return
        end
        
        function [AnnulusPressureResultMD, AnnulusPressureResultVal ] = AnnulusPressureProfile(this)
            AnnulusPressureResultMD = this.AnnulusPressureResultMD;
            AnnulusPressureResultVal = this.AnnulusPressureResultVal;
            return
        end

        function [AnnulusTemperatureResultMD, AnnulusTemperatureResultVal ] = AnnulusTemperatureProfile(this)
            AnnulusTemperatureResultMD = this.AnnulusTemperatureResultMD;
            AnnulusTemperatureResultVal = this.AnnulusTemperatureResultVal;
            return
        end
        
        function GasVolumeFraction = GasVolumeFraction(this,depth)
            idx = find(this.GasVolumeFractionResultMD <= depth , 1);
            
            if idx == 1
                GasVolumeFraction = this.GasVolumeFractionResultVal(1);
                return;
            elseif isempty(idx)
                GasVolumeFraction = this.GasVolumeFractionResultVal(end);
                return;
            else
                depth1 = this.GasVolumeFractionResultMD(idx);
                depth2 = this.GasVolumeFractionResultMD(idx-1);
                
                val1 = this.GasVolumeFractionResultVal(idx);
                val2 = this.GasVolumeFractionResultVal(idx-1);
                
                GasVolumeFraction=interp1([depth1 depth2],[val1 val2],depth);
                return;
            end
        end
        
        %% Stop simulation
        function Stop(this)
            display(['Average step duration: ' num2str(toc(this.TimeAtStep10)/(this.TimeStep-10))]);
            disp('Stopping simulation');
            url = [this.WebServerURL 'simulations/'  this.SimID  '/setpoints' ];
            
            body.Data.FlowRateIn = this.FlowRateIn;
            body.ShouldComplete = true;
            
            [output,extras] = this.readapi(url,'post', jsonencode({body}));
            this.OK = extras.isGood;
            if isfield(output,'Errors')
                error([output.Title output.Detail]);
            end
        end
        
        %% Delete simulation
        function Delete(this)
            disp('Deleting simulation');
            url = [this.WebServerURL 'simulations/'  this.SimID ];
           
            body.Data = [];
            
            [output,extras] = this.readapi(url,'delete', jsonencode({body}));
            this.OK = extras.isGood;
            if isfield(output,'Errors')
                error([output.Title output.Detail]);
            end
        end
        
        %% Get configuration by config guid and version guid (optional)
        function configuration = GetConfiguration(this,config_id, version_id)           
            disp('Getting configuration');
            
            if exist('version_id','var') && ~isempty(version_id)
                url = [this.WebServerURL 'configurations/'  config_id  '/versions/' version_id ];
            else
                url = [this.WebServerURL 'configurations/'  config_id  ];
            end
            
            [output] = this.readapi(url,'get', []);

            outputdecoded = jsondecode(output);
            configuration = outputdecoded.Data;
        end
        
    end
end
