classdef PIcontroller < handle 
% ------------------------------------------------------------------------------------------------------------
% Description: Simple PI controller class
%
% Create PIcontroller object:
%   Constructor: PI = PIcontroller(Kp, Ki, Ts, Reference,InitialChokeOpening); 
%     Kp = Proportional gain
%     Ti = Integral time
%     Ts = Time constant
%     Reference = Reference value / setpoint  in bar 
%     InitialOutput = Initial output
%     ReferenceMaxRateOfChange  = Reference max rate of change in bar 
%     IsReversed = true/false, reversed behaviour of controller, increasing
%     error -> increasing output value
%     Min = min output value
%     Max = max output value
%
% Methods: 
%   Reset( Reference, InitialOutput) - Reset controller
%     Reference = Reference value / setpoint
%     InitialOutput = Output value to initialize to
%
%   Output = GetOutput( Measured ) - Get output value      
%     Measured - Measure value
%     Output - Output value
% 
% Created: 2023-11-04 by Sonja Moi
% History:
% -------------------------------------------------------------------------
    
    properties (Access = public)
        Kp;
        Ti;
        Ts;
        Reference;
        ReferenceMaxRateOfChange;
    end
    
    properties (Access = private)
        Error = 0;
        Proportional = 0;
        Integral = 0;
        ErrorSum = 0;
        ReferenceLast;
        IsReversed;
        Min;
        Max;        
        ResetStatus = true;
        Measured;
        Output;
    end
    
    methods
        function this = PIcontrollerV2(Kp, Ti, Ts, Reference, InitialOutput, ReferenceMaxRateOfChange,IsReversed, Min, Max)
            this.Kp = Kp;
            this.Ti = Ti;
            this.Ts = Ts;
            this.Reference = Reference;
            this.ReferenceMaxRateOfChange = ReferenceMaxRateOfChange;
            this.Min = Min;
            this.Max = Max;
            
            if IsReversed
                this.IsReversed = -1; 
            else
                this.IsReversed = 1; 
            end
        
            this.ErrorSum = InitialOutput / (this.Kp / this.Ti) * this.IsReversed;
        end
        
        function this = SoftReset(this)
            this.ResetStatus = true;
            this.ReferenceLast = this.Reference;
            this.ErrorSum = this.Output / (this.Kp / this.Ti) * this.IsReversed;
        end
        
        function this = Reset(this, Reference, Measured)
            this.ResetStatus = true;
            this.Reference = Reference;
            this.ReferenceLast = Reference;
            this.Output = Measured;
            this.ErrorSum = Measured / (this.Kp / this.Ti) * this.IsReversed;
        end

        function value = RateLimit(this, Reference)
            delta = Reference - this.ReferenceLast;
            deltasign = sign(delta);
            value = this.ReferenceLast + (deltasign * min(abs(delta), this.ReferenceMaxRateOfChange * this.Ts)); 
        end
        
        function Output = GetOutput(this, Reference, Measured)
            this.Measured = Measured;
            this.Reference = this.RateLimit(Reference);
            if this.ResetStatus
                this.Error=(this.Reference-this.Measured);
                this.ReferenceLast = this.Reference;                
                this.ResetStatus = false;
            else
                this.Error = this.Reference - Measured;
            end             
            
            this.Proportional = this.Kp * this.IsReversed * this.Error;
            this.ErrorSum = this.ErrorSum  + (this.Error * this.Ts);
            this.Integral = (this.Kp / this.Ti) * this.IsReversed * this.ErrorSum;
            
            this.Output = this.Proportional + this.Integral;            
            
            if this.Output>this.Max
                this.Output=this.Max;
                this.Integral = this.Output;
                this.ErrorSum = this.Output / (this.Kp / this.Ti) * this.IsReversed;
            end
            if this.Output<this.Min
                this.Output=this.Min;                
                this.Integral = this.Output;
                this.ErrorSum = this.Output / (this.Kp / this.Ti) * this.IsReversed;
            end
           
            this.ReferenceLast = this.Reference;
            Output = this.Output;
        end  
    end  
end