clear variables;
clc;
close all;

% Add to the matlab path the folder where this file is located and all the
% subfolders
addpath(genpath(pwd));

IdentityServerURL = 'https://live.openlab.app/';

[username, api_key, license_guid] = GetLoginData();

ConfigurationName = 'influx';
SimulationName = 'Manual influx (Matlab)';
InitialBitDepth = 2497; % [m]
InitialTopOfStringPosition = 20;

% Influx/loss configuration
UseReservoirModel = true; % false: no influx and loss, true: influx and loss
ManualReservoirMode = false; % Influx based on geo-thermal (only if UseReservoirModel is true)
ReservoirKickOffTime = 15; % [s]

%% Initialization
% Create simulation object
Sim = OpenLabClient(IdentityServerURL,username,api_key,license_guid,ConfigurationName,SimulationName,InitialBitDepth,InitialTopOfStringPosition,...
    UseReservoirModel,ManualReservoirMode,0,0,0,ReservoirKickOffTime);

%% Simulation
% Initial setpoints
Sim.TopOfStringVelocity = 0;
Sim.SurfaceRPM = 0;
Sim.ChokePumpFlowRateIn = 0;
Sim.ChokeOpening = 1;
Sim.BopChokeOpening = 1;
Sim.FlowRateIn = 250/60000;
Sim.MainPitMainPump = true; % Pump from main pit
Sim.MainPitReturn = true; % Return mud to main pit
Sim.MainPitKillPump = true; % Pump from main pit

if Sim.IsOK
    try
        for timeStep = 1 : 12000
            tStartStep = tic;           
            
            % Close BOP and reduce rig choke opening. This will increase
            % the bottom hole pressure and the influx will stop. Continue 
            % pumping until the kick is circulated out.
            % Check the reservoir flow plot in the web client to see if 
            % there is an influx.
            % Check the gas volume depth profile plot in the web client to
            % monitor when the influx is circulated out of the well.
            if timeStep > 60
                Sim.BopChokeOpening = 0;
                Sim.ChokelineValve = 1;
                Sim.RigChokeOpening = 0.025;
                Sim.KillineValve = 0;                  
                Sim.KillPumpFlowRateIn = 0/60000;
                Sim.MainPitKillPump = true;
                Sim.MainPitMainPump = true; % Pump from main pit
                Sim.MainPitReturn = true; % Return mud to main pit                
                Sim.FlowRateIn = 250/60000;
            end          
            
            % If the drillstring is above the bop, or not present in the
            % drill hole, then the kill pump must be used to increase the
            % bottom hole pressue and stop the influx.
%             if timeStep > 60
%                 Sim.BopChokeOpening = 0;
%                 Sim.ChokelineValve = 1;
%                 Sim.RigChokeOpening = 0.025;
%                 Sim.KillineValve = 1;
%                 Sim.KillPumpFlowRateIn = 500/60000;
%                 Sim.MainPitKillPump = true;
%                 Sim.MainPitMainPump = true;
%                 Sim.MainPitReturn = true;
%                 Sim.FlowRateIn = 0/60000;
%             end
            
            % Check the density depth profile plot in the web client to
            % monitor the progress of pumping heavyer fluid from reserve
            % tank.
            if timeStep > 5300                
                Sim.BopChokeOpening = 0;
                Sim.ChokelineValve = 1;
                Sim.RigChokeOpening = 0.025;
                Sim.KillineValve = 0;                  
                Sim.KillPumpFlowRateIn = 250/60000;
                Sim.MainPitKillPump = true;
                Sim.MainPitMainPump = false; % Pump from reserve pit
                Sim.MainPitReturn = true; % Pump to main pit (must switch to reserve pit when new mud reaches the outlet of the well.               
                Sim.FlowRateIn = 250/60000;
            end
            
            Sim.Step();
            
            if ~Sim.IsOK
                break;
            end
                        
            display(['Total step duration: ' num2str(toc(tStartStep))]);
        end
        
        % Stop Simulation process
        Sim.Stop;
        
    catch exception
        % Stop simulation
        Sim.Stop;
        
        rethrow(exception);
    end
    
end
